#include <gint/keyboard.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "string.h"
#include <gint/mpu/rtc.h>
#pragma once

#ifndef ClassParticuleEngine
#define ClassParticuleEngine

extern "C"


#include "Announcement.h"
#include "ParticuleGraphics.hpp"
#include "List.h"

#include <math.h>

#include "Ressources.h"
#include "ParticuleBase.hpp"



int Random(int start, int end);

int LenChar(char* txt);

int mod(int x, int m);

template <class Event>
struct ParticuleEvent {
public:
    void (Event::* function)();
    Event* self;

    void Set(void (Event::* func)(), Event* obj) {
        function = func;
        self = obj;
    }

    void Invoke()
    {
        if (self!=NULL && function!=NULL)
            (self->*function)();
    };
};


enum Tag
{
    Untagged = 0,
    Respawn = 1,
    Finish = 2,
    EditorOnly = 3,
    MainCamera = 4,
    Player = 5,
    GameController = 6
};

enum Layer
{
    Default = 0,
    TransparentFX = 1,
    IgnoreRaycast = 2,
    Water = 3,
    UI = 4
};





class Component : public Object {
    //https://docs.unity3d.com/ScriptReference/Component.html
public:
    //char* tag;
    GameObject* gameObject;
    //Transform* transform;

    Component(const char* name, GameObject* gameObject, const char* UUID = NULL);

    virtual ~Component();

    /*void BroadcastMessage();
    void CompareTag();
    void GetComponent();
    void GetComponentInChildren();
    void GetComponentInParent();
    void GetComponents();
    void GetComponentsInChildren();
    void GetComponentsInParent();
    void SendMessage();
    void SendMessageUpwards();
    void TryGetComponent();*/

    ///
    virtual void PhysicsCalculator() {};
    ///

    virtual void Awake() {}
    virtual void FixedUpdate() {}
    virtual void LateUpdate() {}
    /*virtual void OnAnimatorIK() {}
    virtual void OnAnimatorMove() {}
    virtual void OnApplicationFocus() {}
    virtual void OnApplicationPause() {}
    virtual void OnApplicationQuit() {}
    virtual void OnAudioFilterRead() {}
    virtual void OnBecameInvisible() {}
    virtual void OnBecameVisible() {}*/
    //virtual void OnCollisionEnter() {}
    virtual void OnCollisionEnter2D(Collider2D* collider2D) {}
    //virtual void OnCollisionExit() {}
    virtual void OnCollisionExit2D(Collider2D* collider2D) {}
    //virtual void OnCollisionStay() {}
    virtual void OnCollisionStay2D(Collider2D* collider2D) {}
    //virtual void OnConnectedToServer() {}
    virtual void OnControllerColliderHit() {}//////
    virtual void OnDestroy() {}
    void Destroy() {
        OnDestroy();
        delete this;
    }
    //virtual void OnDisable() {}//////
    /*virtual void OnDisconnectedFromServer() {}
    virtual void OnDrawGizmos() {}
    virtual void OnDrawGizmosSelected() {}*/
    //virtual void OnEnable() {}//////
    //virtual void OnFailedToConnect() {}
    //virtual void OnFailedToConnectToMasterServer() {}
    virtual void OnGUI() {}///////
    /*virtual void OnJointBreak() {}
    virtual void OnJointBreak2D() {}
    virtual void OnMasterServerEvent() {}
    virtual void OnMouseDown() {}
    virtual void OnMouseDrag() {}
    virtual void OnMouseEnter() {}
    virtual void OnMouseExit() {}
    virtual void OnMouseOver() {}
    virtual void OnMouseUp() {}
    virtual void OnMouseUpAsButton() {}
    virtual void OnNetworkInstantiate() {}
    virtual void OnParticleCollision() {}
    virtual void OnParticleSystemStopped() {}
    virtual void OnParticleTrigger() {}
    virtual void OnParticleUpdateJobScheduled() {}
    virtual void OnPlayerConnected() {}
    virtual void OnPlayerDisconnected() {}
    virtual void OnPostRender() {}
    virtual void OnPreCull() {}
    virtual void OnPreRender() {}
    virtual void OnRenderImage() {}*/
    virtual void OnRenderObject() {}
    /*virtual void OnSerializeNetworkView() {}
    virtual void OnServerInitialized() {}
    virtual void OnTransformChildrenChanged() {}
    virtual void OnTransformParentChanged() {}*/
    //virtual void OnTriggerEnter() {}
    virtual void OnTriggerEnter2D(Collider2D* collider2D) {}
    //virtual void OnTriggerExit() {}
    virtual void OnTriggerExit2D(Collider2D* collider2D) {}
    //virtual void OnTriggerStay() {}
    virtual void OnTriggerStay2D(Collider2D* collider2D) {}
    virtual void OnValidate() {}//////
    //virtual void OnWillRenderObject() {}
    //virtual void Reset() {}/////
    virtual void Start() {}
    virtual void Update() {}
};
class Transform : public Component {
    //https://docs.unity3d.com/ScriptReference/Transform.html
private:
    Vector2* lastPosition;
public:
    List<Transform*> children;
    int childCount;
    //Vector2* eulerAngles;
    //Vector2* forward;
    bool hasChanged;
    int hierarchyCapacity;
    int hierarchyCount;
    //Vector2* localEulerAngles;
    Vector2* localPosition;
    //Quaternion* localRotation;
    //Vector2* localScale;
    //Matrix4x4* localToWorldMatrix;
    //Vector2* lossyScale;
    Transform* parent;
    Vector2* position;
    //Vector2* right;
    //Transform* root;
    //Quaternion* rotation;
    //Vector2* up;
    //Matrix4x4* worldToLocalMatrix;

    Transform(GameObject* gameObject, const char* UUID = NULL, const char* name = "Transform");

    ~Transform() {
        delete lastPosition;
        delete localPosition;
        //delete localScale;
        delete position;
    };

    void SetParent(Transform* transform);

    void Start();

    void Update();
};


class Behaviour : public Component {
    //https://docs.unity3d.com/ScriptReference/Behaviour.html
public:

    bool enabled;
    bool isActiveAndEnabled;
    Behaviour(const char* name, GameObject* gameObject, const char* UUID = NULL);

    virtual ~Behaviour() {};
};

/*
enum ScaleMode
{
    //https://docs.unity3d.com/ScriptReference/ScaleMode.html
    StretchToFill,
    ScaleAndCrop,
    ScaleToFit,
};
class CallbackEventHandler{
    //https://docs.unity3d.com/ScriptReference/UIElements.CallbackEventHandler.html

};

class Focusable : public CallbackEventHandler {
    //https://docs.unity3d.com/ScriptReference/UIElements.Focusable.html

};

class VisualElement : public Focusable {
    //https://docs.unity3d.com/ScriptReference/UIElements.VisualElement.html

};
*/




class MonoBehaviour : public Behaviour {
    //https://docs.unity3d.com/ScriptReference/MonoBehaviour.html
public:
    //runInEditMode c'est pour l'diteur donc il n'est pas prsent
    bool useGUILayout;

    MonoBehaviour(const char* name, GameObject* gameObject, const char* UUID = NULL);

    virtual ~MonoBehaviour() {};

    /*void CancelInvoke();
    void Invoke();
    void InvokeRepeating();
    void IsInvoking();
    void StartCoroutine();
    void StopAllCoroutines();
    void StopCoroutine();*/
    //print c'est pour l'diteur donc il n'est pas prsent
};


class GameObject : public Object {
    //https://docs.unity3d.com/ScriptReference/GameObject.html
public:
    List<Component*> ListOfComponent;
    bool activeInHierarchy;
    bool activeSelf;
    bool isStatic;
    Layer layer;
    Scene* scene;
    long sceneCullingMask;
    Tag tag;
    Transform* transform;

    GameObject(Scene* scene, const char* name, const char* UUID = NULL);

    ~GameObject();

    bool IsActive();

    void AddComponent(Component* component);
    //void BroadcastMessage();
    bool CompareTag(Tag tag);
    Component* GetComponent(const char* name);
    /*void GetComponentInChildren();
    void GetComponentInParent();
    void GetComponents();
    void GetComponentsInChildren();
    void GetComponentsInParent();
    void SendMessage();
    void SendMessageUpwards();
    void SetActive();
    void TryGetComponent();*/
    void Destroy() {
        delete this;
    }

    //static void CreatePrimitive();
    GameObject* Find(unsigned char* name);
    //static void FindGameObjectsWithTag();
    GameObject* FindWithTag(Tag tag);
};


class Scene {
    //https://docs.unity3d.com/ScriptReference/SceneManagement.Scene.html
public:
    List<GameObject*> AllGameObject;
    SceneManager* sceneManager;
    List<Camera*> AllCameras;
    List<Collider2D*> LstColliders;
    //void GetRootGameObjects();
    //bool IsValid();

    ~Scene() {
        for (int i = 0; i < AllGameObject.Count; i++) {
            //delete AllGameObject.Pop();
            AllGameObject[0]->Destroy();
        }
        AllGameObject.Clear();
    }

    void AddGameObject(GameObject* gObj);

    void Update();

    void Start();

    int buildIndex;
    //bool isDirty;
    //bool isLoaded;
    char* name;
    //char* path;
    //int rootCount;

    
};

class Camera : public Behaviour {
    //https://docs.unity3d.com/ScriptReference/Camera.html
public:
    Camera(GameObject* gameObject, const char* UUID = NULL);

    void OnDestroy();
    void SetMainCamera();
};

class Image : public MonoBehaviour { //: public VisualElement {
    //https://docs.unity3d.com/ScriptReference/UIElements.Image.html
public:
    //unsigned char* ussClassName;
    Texture* image;
    //scaleMode
    //sourceRect
    //tintColor
    //uv
    //vectorImage

    Image(GameObject* gameObject, Texture* image, const char* UUID = NULL);

    void OnRenderObject();
};

class Sprite : public MonoBehaviour { //: public VisualElement {
    //https://docs.unity3d.com/ScriptReference/UIElements.Image.html
public:

    Texture* image;
    bool HaveBackground;


    Sprite(GameObject* gameObject, Texture* image,bool HaveBackground, const char* UUID = NULL);

    void OnRenderObject();
};





class Collider2D : public MonoBehaviour {
    //https://docs.unity3d.com/ScriptReference/Collider2D.html
protected:
    List<Collider2D*> LstColliders;
public:
    bool IsTrigger;
    Collider2D(GameObject* gameObject, bool IsTrigger, const char* UUID = NULL) : MonoBehaviour("Collider2D", gameObject, UUID) {
        this->IsTrigger = IsTrigger;
    };

    void Awake() {
        gameObject->scene->LstColliders.Add(this);
    }

    void Update();

    bool Contains(Collider2D* collider) {
        for (int o = 0; o < LstColliders.Count; o++) {
            if (collider == LstColliders[o])
                return true;

        }
        return false;
    };

    virtual bool AreTheyTouching(Collider2D* boxCollider) { return false; };
    
};

class BoxCollider2D : public Collider2D {//Collider2D {
    //https://docs.unity3d.com/ScriptReference/BoxCollider2D.html


public:
    Vector2* center;
    Vector2* size;

    BoxCollider2D(GameObject* gameObject,bool IsTrigger,Vector2* center,Vector2* size, const char* UUID = NULL);

    

    bool AreTheyTouching(Collider2D* boxCollider);

    

    ~BoxCollider2D() {
        delete center;
        delete size;
    };
};

class Rigidbody : public MonoBehaviour {
private:
    Collider2D* MyCollider;
    Vector2* lastPosition;

    bool CheckCollider();
    bool IsVisible();
public:
    float Mass;
    bool UseGravity;
    bool IsKinematic;
    Vector2* velocity;

    Rigidbody(GameObject* gameObject, float Mass, bool UseGravity, bool IsKinematic, const char* UUID = NULL);

    void PhysicsCalculator();


    ~Rigidbody() {
        delete velocity;
        delete lastPosition;
    };

    void Start();
    void LateUpdate();

};

class Text : public MonoBehaviour {

public:
    unsigned char* text;

    Text(GameObject* gameObject, unsigned char* text, const char* UUID = NULL);

    void OnRenderObject();


};

class Tilemap : public MonoBehaviour {
public:

    Texture** images;
    int* Datas;
    Vector2* sizeTilemap;
    Vector2* sizeCase;


    Tilemap(GameObject* gameObject, Vector2* sizeTilemap, Vector2* sizeCase, const char* UUID = NULL);

    ~Tilemap() {
        delete[] images;
        delete[] Datas;
    };

    void OnRenderObject();
};

class CanvasUI : public MonoBehaviour {
    bool wait;
public:
    List<Component*> AllChildrenUI;
    Component* ButtonSelection;
    CanvasUI(GameObject* gameObject, const char* UUID = NULL);
    ~CanvasUI() {
        AllChildrenUI.Clear();
    };
    void Update();
    void FixedUpdate();
    void OnRenderObject();
};

template <class T>
class ButtonUI : public MonoBehaviour {
public:
    //void (*OnClick)();
    //OnClick = (&CanvasUI::Move);
    //(this->*OnClick)();
    Texture* NormalTexture;
    Texture* HighlightedTexture;

    ParticuleEvent<T> OnClick;
    CanvasUI* canvas;

    Sprite* sprite;


    ButtonUI(GameObject* gameObject, Texture* NormalTexture, Texture* HighlightedTexture, const char* UUID = NULL) : MonoBehaviour("ButtonUI", gameObject, UUID) {
        this->NormalTexture = NormalTexture;
        this->HighlightedTexture = HighlightedTexture;
    };

    ~ButtonUI() {
        if (canvas != NULL) {
            if (canvas->AllChildrenUI.Contains(this)) {
                canvas->AllChildrenUI.Remove(this);
            }
        }
    };

    void Update() {
        if (canvas != NULL && canvas->ButtonSelection == this) {
            if (IsKeyDown(KEY_CTRL_EXE))
                OnClick.Invoke();
        }
    };

    void Awake() {
        if (sprite == NULL) {
            sprite = (Sprite*)gameObject->GetComponent("Sprite");
        }
        else {
            if (NormalTexture == NULL) {
                NormalTexture = sprite->image;
            }
        }

        Transform* tempParent = this->gameObject->transform;
        while (tempParent != NULL && this->canvas == NULL)
        {
            this->canvas = ((CanvasUI*)tempParent->gameObject->GetComponent("CanvasUI"));
            if (this->canvas == NULL)
            {
                tempParent = tempParent->parent;
            }
        }
        if (this->canvas != NULL) {
            canvas->AllChildrenUI.Add(this);
        }
    }

    void LateUpdate() {

        if (sprite != NULL) {
            sprite->HaveBackground = true;
            if (canvas != NULL && canvas->ButtonSelection == this) {

                if (sprite->image != HighlightedTexture) {
                    sprite->image = HighlightedTexture;
                }
                DrawRectangle(gameObject->transform->localPosition->x - 1, gameObject->transform->localPosition->y - 1, sprite->image->width + 2, sprite->image->height + 2, ML_BLACK);
            }
            else
            {

                if (sprite->image != NormalTexture) {
                    sprite->image = NormalTexture;
                }
            }
        }

    };
};


class ChangeScene :  public MonoBehaviour {
private:
    
public:
    
    ChangeScene(GameObject* gameObject , const char* UUID = NULL) : MonoBehaviour("ChangeScene", gameObject, UUID) {
        
    };
    
    
};
class TestSc :  public MonoBehaviour {
private:
    
public:
    
    TestSc(GameObject* gameObject , const char* UUID = NULL) : MonoBehaviour("TestSc", gameObject, UUID) {
        
    };
    
    
};

class ProjectSettings {
public:
    Vector2* Gravity;
    Vector2* ScreenSize;

    
    ProjectSettings() {
        //<ProjectSettings>
        Gravity = new Vector2(0.0f,0.01f);
        ScreenSize = new Vector2(396,224);
        //<\ProjectSettings>
    };

};
class SceneManager {
    //https://docs.unity3d.com/ScriptReference/SceneManagement.SceneManager.html
public:
    int sceneCount;
    static int sceneCountInBuildSettings;
    int LoadSceneAfter;
    bool _quit;
    ProjectSettings* projectSettings;
    
    SceneManager() {
        LoadSceneAfter = -1;
        _quit = false;
        projectSettings = new ProjectSettings();
        LoadTextures();
    }

    //Static Methods
    /*static void CreateScene();
    static void GetActiveScene();
    static void GetSceneAt();
    static void GetSceneByBuildIndex();
    static void GetSceneByName();
    static void GetSceneByPath();*/
    void LoadScene(int index);
    void LoadScene(unsigned char* name);
    /*static void LoadSceneAsync();
    static void MergeScenes();
    static void MoveGameObjectToScene();
    static void SetActiveScene();
    static void UnloadSceneAsync();*/

    GameObject* FindWithName(unsigned char* name);

    //Events
    /*static void activeSceneChanged();
    static void sceneLoaded();
    static void sceneUnloaded();*/

    void UpdateScene();

    void StartScene();

    Texture* Texture_TextureVide;
    Texture* Texture_UUID_4393f5ea_ebf4_40af_868d_e0e560bd4069;
Texture* Texture_UUID_e4b4a56f_c79f_447c_bcec_a678a6223596;


private:
    List<Scene*> AllSceneLoaded;

    void LoadTextures() {
        Texture_TextureVide = new Texture();
        //CreateTextures
    };


    Scene* GetSceneInBuild(int index) {
        Scene* newScene = new Scene();
        newScene->sceneManager = this;

        if (index == 0){
GameObject* UUID_978ce220_6e3e_4a21_8e90_e511492338e5;
GameObject* UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878;
GameObject* UUID_c631094b_1221_42d2_9f4c_758d9e97ca61;
GameObject* UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b;
Sprite* UUID_0ab9893a_e320_4176_bd82_b1d93f4c5357;

ButtonUI<Object>* UUID_28198b47_8d32_47c6_9320_e11c34fa1605;
Sprite* UUID_0bc3c15d_2735_4d6d_97a6_5ffe0d2bd49d;

CanvasUI* UUID_0130794b_4ede_46f4_960a_08f11be26a15;

Camera* UUID_23480279_970a_4275_bb0f_75b1b5b2d79a;

UUID_978ce220_6e3e_4a21_8e90_e511492338e5= new GameObject(newScene, "Camera","UUID_978ce220_6e3e_4a21_8e90_e511492338e5");
UUID_978ce220_6e3e_4a21_8e90_e511492338e5->activeSelf = true;
UUID_978ce220_6e3e_4a21_8e90_e511492338e5->activeInHierarchy = true;
UUID_978ce220_6e3e_4a21_8e90_e511492338e5->isStatic = false;
UUID_978ce220_6e3e_4a21_8e90_e511492338e5->transform->position->Set(0,0);
UUID_978ce220_6e3e_4a21_8e90_e511492338e5->transform->localPosition->Set(0,0);
newScene->AddGameObject(UUID_978ce220_6e3e_4a21_8e90_e511492338e5);
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878= new GameObject(newScene, "Canvas","UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878");
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->activeSelf = true;
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->activeInHierarchy = true;
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->isStatic = false;
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->transform->position->Set(428,3);
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->transform->localPosition->Set(428,3);
newScene->AddGameObject(UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878);
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61= new GameObject(newScene, "GameObject (1)","UUID_c631094b_1221_42d2_9f4c_758d9e97ca61");
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->activeSelf = true;
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->activeInHierarchy = true;
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->isStatic = false;
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->transform->position->Set(461.0,102.0);
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->transform->localPosition->Set(33.0,99.0);
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->transform->SetParent(UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->transform);newScene->AddGameObject(UUID_c631094b_1221_42d2_9f4c_758d9e97ca61);
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b= new GameObject(newScene, "GameObject","UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b");
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b->activeSelf = true;
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b->activeInHierarchy = true;
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b->isStatic = false;
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b->transform->position->Set(47.0,17.0);
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b->transform->localPosition->Set(47.0,17.0);
newScene->AddGameObject(UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b);




UUID_23480279_970a_4275_bb0f_75b1b5b2d79a = new Camera(UUID_978ce220_6e3e_4a21_8e90_e511492338e5,"UUID_23480279_970a_4275_bb0f_75b1b5b2d79a");
UUID_978ce220_6e3e_4a21_8e90_e511492338e5->AddComponent((Component*)UUID_23480279_970a_4275_bb0f_75b1b5b2d79a);





UUID_0130794b_4ede_46f4_960a_08f11be26a15 = new CanvasUI(UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878,"UUID_0130794b_4ede_46f4_960a_08f11be26a15");
UUID_e2b3119e_c7a9_449c_8132_c372cd5a3878->AddComponent((Component*)UUID_0130794b_4ede_46f4_960a_08f11be26a15);





UUID_0bc3c15d_2735_4d6d_97a6_5ffe0d2bd49d = new Sprite(UUID_c631094b_1221_42d2_9f4c_758d9e97ca61,Texture_UUID_4393f5ea_ebf4_40af_868d_e0e560bd4069,false,"UUID_0bc3c15d_2735_4d6d_97a6_5ffe0d2bd49d");
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->AddComponent((Component*)UUID_0bc3c15d_2735_4d6d_97a6_5ffe0d2bd49d);


UUID_28198b47_8d32_47c6_9320_e11c34fa1605 = new ButtonUI<Object>(UUID_c631094b_1221_42d2_9f4c_758d9e97ca61,Texture_UUID_e4b4a56f_c79f_447c_bcec_a678a6223596,Texture_UUID_4393f5ea_ebf4_40af_868d_e0e560bd4069,"UUID_28198b47_8d32_47c6_9320_e11c34fa1605");
UUID_c631094b_1221_42d2_9f4c_758d9e97ca61->AddComponent((Component*)UUID_28198b47_8d32_47c6_9320_e11c34fa1605);





UUID_0ab9893a_e320_4176_bd82_b1d93f4c5357 = new Sprite(UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b,Texture_UUID_4393f5ea_ebf4_40af_868d_e0e560bd4069,false,"UUID_0ab9893a_e320_4176_bd82_b1d93f4c5357");
UUID_d22ebb64_acdd_4b0a_bf27_37f626976c3b->AddComponent((Component*)UUID_0ab9893a_e320_4176_bd82_b1d93f4c5357);


}

        return newScene;

    }

    Scene* GetSceneInBuild(unsigned char* name) {
        unsigned char* AllName[] = { (unsigned char*)"Main" };
        int count = 0;
        for (int i = 0; i < count; i++)
        {
            if (name == AllName[i]) {
                return GetSceneInBuild(i);
            }
        }
        return NULL;
    }
};


void LoadScene(Scene* scene, int nb);

GameObject* FindWithName(unsigned char* name, GameObject* gameObject);

#endif